/* Construction primitive operations on 2D tessellations (Euler's operators) */
#include "dcel.h"
#include "circular.h"
#include "relation.h"
#include "geometry.h"

#ifndef TRUE
typedef int bool;
#define TRUE 1
#define FALSE 0
#endif

#ifndef malloc
char *malloc();
#endif

#define new(t) (t *)malloc(sizeof(t))


/* Low-level functions */

static bool new_vertex(xx,yy,v)
double xx,yy; vptr *v;
{
 if (((*v)=new(vertex))==NULL)
  return(FALSE);
 (*v)->x=xx;
 (*v)->y=yy;
 (*v)->incidentedge=NULL;
 (*v)->mark=FALSE;
 (*v)->next=NULL;
 (*v)->prev=NULL;
 return(TRUE);
}

static bool new_face(f)
fptr *f;
{
 if (((*f)=new(face))==NULL)
  return(FALSE);
 (*f)->boundingedge=NULL;
 (*f)->next=NULL;
 (*f)->prev=NULL;
 (*f)->mark=FALSE;
 (*f)->xc=0;
 (*f)->yc=0;
 (*f)->rc=0;
 return(TRUE);
}

static bool new_edge(s,e)
sptr s; eptr *e;
{
 if (((*e)=new(edge))==NULL)
  return(FALSE);
 (*e)->mark=FALSE;
 (*e)->constraint=FALSE;
 (*e)->vertices[0]=NULL;
 (*e)->vertices[1]=NULL;
 (*e)->faces[0]=NULL;
 (*e)->faces[1]=NULL;
 (*e)->next=NULL;
 (*e)->prev=NULL;
 (*e)->structure=s;
 return(TRUE);
}

/* Construction primitive operations (Euler's operators) */

bool MAKE_VERTEX_INFINITY_FACE(s,x,y,v,f)
/* initializes surface s by inserting vertex v of
   coordinates (x,y) and creating infinity face f */
sptr *s; double x,y; vptr *v; fptr *f;
{
 if (((*s)=new(surface))==NULL) return(FALSE);
 if (!new_vertex(x,y,v)) return(FALSE);
 if (!new_face(f)) return(FALSE);
 (*s)->vlist=(*v);
 (*s)->elist=NULL;
 (*s)->flist=(*f);
 (*s)->infinity=(*f);
 return(TRUE);
}

static fptr get_face(s,e,e1,v)
sptr s; eptr e,e1; vptr v; 
{
 fptr f;double xv,yv;

 if ((e->faces[0]==e1->faces[0]) ||    /* find a face common */
     (e->faces[0]==e1->faces[1]))      /* to e and e1        */
  f=e->faces[0];
 else if ((e->faces[1]==e1->faces[0]) ||
          (e->faces[1]==e1->faces[1]))
  f=e->faces[1];
 COORDINATES(v,&xv,&yv);
 if (ENCLOSES(s,f,xv,yv))
  return(f);
 else
  return(e->faces[f==e->faces[0]]);
}

bool MAKE_EDGE_VERTEX(s,v,x,y,vn,e)
sptr s; vptr v,*vn; eptr *e; double x,y;
/* creates a new edge e in s, connecting vertex v of s with
   new vertex vn of coordinates (x,y) */
{
 lptr el; fptr f;eptr e0,e1;

 if (!IS_A_VERTEX(s,v)) return(FALSE);
 if (!new_vertex(x,y,vn)) return(FALSE);
 if (!new_edge(s,e)) return(FALSE);
 if (IS_ISOLATED(s,v))                 /* first edge */
 {
  (*e)->edges[0]=(*e);
  v->incidentedge=(*e);
  f=s->infinity;
  f->boundingedge=(*e);
 }
 else if (IS_DANGLING_VERTEX(s,v))     /* dangling chain */
 {
  (*e)->edges[0]=v->incidentedge;
  v->incidentedge->edges[v==v->incidentedge->vertices[1]]=(*e);
  f=v->incidentedge->faces[0];
 }
 else                                  /* regular case */
 {
  el=VERTEX_EDGE(s,v);
  SECTOR(v,el,*vn,&e0,&e1);
  f=get_face(s,e1,e0,*vn);
  e1->edges[v==e1->vertices[1]]=(*e);
  (*e)->edges[0]=e0;
  FREELIST(&el);
 }
 (*vn)->incidentedge=(*e);           /* set remaining links */
 (*e)->vertices[0]=v;
 (*e)->vertices[1]=(*vn);
 (*e)->faces[0]=f;
 (*e)->faces[1]=f;
 (*e)->edges[1]=(*e);
 (*vn)->next=s->vlist;
 s->vlist->prev=(*vn);
 s->vlist=(*vn);
 (*e)->next=s->elist;
 if (s->elist!=NULL) 
    s->elist->prev=(*e);
 s->elist=(*e);
 return(TRUE);
}

bool MAKE_EDGE_FACE(s,fo,fn,v0,v1,e)
sptr s; fptr fo,*fn; vptr v0,v1; eptr *e;
/* splits face fo of s by a new edge e connecting vertices v0
   and v1; a new face fn is created to the left of vector v1-v0 */
/* WARNING: when splitting infinity face, v0 and v1 must be ordered
   so that fo corresponds to infinity face also after the operation */
{
 lptr el0,el1; eptr ek,ek1,eh,eh1; vptr vh;

 if (!(IS_A_FACE(s,fo) && IS_A_VERTEX(s,v0) && IS_A_VERTEX(s,v1)))
  return(FALSE);
 if (!new_face(fn)) return(FALSE);
 if (!new_edge(s,e)) return(FALSE);
 el0=VERTEX_EDGE(s,v0);
 el1=VERTEX_EDGE(s,v1);
 SECTOR(v1,el1,v0,&ek,&ek1);
 SECTOR(v0,el0,v1,&eh,&eh1);
 (*e)->edges[0]=eh;                    /* set edge links of e */
 (*e)->edges[1]=ek;
 vh=v0;
 do                    /* set face links of edges bounding fn */
 {
  eh->faces[vh==eh->vertices[0]]=(*fn);
  vh=eh->vertices[vh==eh->vertices[0]];
  eh=eh->edges[vh==eh->vertices[1]];
 } while ((eh!=ek) || (vh!=v1));
                       /* set edge links of edges adjacent to e */
 ek1->edges[v1==ek1->vertices[1]]=(*e);
 eh1->edges[v0==eh1->vertices[1]]=(*e);
 fo->boundingedge=(*e);                  /* set remaining links */
 fo->rc=0;
 (*fn)->boundingedge=(*e);
 (*e)->vertices[0]=v0;
 (*e)->vertices[1]=v1;
 (*e)->faces[0]=(*fn);
 (*e)->faces[1]=fo;
 (*fn)->next=s->flist;               /* add face and edge to list */
 s->flist->prev=(*fn);
 s->flist=(*fn);
 (*e)->next=s->elist;
 if (s->elist!=NULL) 
  s->elist->prev=(*e);
 s->elist=(*e);
 FREELIST(&el0);
 FREELIST(&el1);
 return(TRUE);
}

bool MAKE_EDGE_BOUND_FACE(s,fo,fn,v0,v1,e,ek)
sptr s; fptr fo,*fn; vptr v0,v1; eptr *e,*ek;
/* splits face fo of s by a new edge e connecting vertices v0
   and v1; a new face fn is created to the left of vector v1-v0 */
/* WARNING: when splitting infinity face, v0 and v1 must be ordered
   so that fo corresponds to infinity face also after the operation */
{
 lptr el0; eptr eh,eh1; vptr vh;

 if (!(IS_A_FACE(s,fo) && IS_A_VERTEX(s,v0) && IS_A_VERTEX(s,v1)))
  return(FALSE);
 if (!new_face(fn)) return(FALSE);
 if (!new_edge(s,e)) return(FALSE);
 el0=VERTEX_EDGE(s,v0);
 SECTOR(v0,el0,v1,&eh,&eh1);
 (*e)->edges[0]=eh;                    /* set edge links of e */
 (*e)->edges[1]=(*ek);
 vh=v0;
                       /* set face links of edges bounding fn */
 while (vh!=((*ek)->vertices[v1==(*ek)->vertices[0]]))
 {
  eh->faces[vh==eh->vertices[0]]=(*fn);
  vh=eh->vertices[vh==eh->vertices[0]];
  eh=eh->edges[vh==eh->vertices[1]];
 }
 (*ek)->faces[v1==(*ek)->vertices[0]]=fo;
 (*ek)->faces[v1==(*ek)->vertices[1]]=(*fn);
                       /* set edge links of edges adjacent to e */
 eh1->edges[v0==eh1->vertices[1]]=(*e);
 (*ek)->edges[v1==(*ek)->vertices[1]]=(*e);
 fo->boundingedge=(*e);                  /* set remaining links */
 fo->rc=0;
 (*fn)->boundingedge=(*e);
 (*e)->vertices[0]=v0;
 (*e)->vertices[1]=v1;
 (*e)->faces[0]=(*fn);
 (*e)->faces[1]=fo;
 v1->incidentedge=(*e);
 (*fn)->next=s->flist;               /* add face and edge to list */
 s->flist->prev=(*fn);
 s->flist=(*fn);
 (*e)->next=s->elist;
 if (s->elist!=NULL) 
  s->elist->prev=(*e);
 s->elist=(*e);
 FREELIST(&el0);
 return(TRUE);
}

bool KILL_EDGE_FACE(s,e,f)
sptr s; eptr *e; fptr *f;
/* deletes face f and edge e from tassellation s */
{
 fptr f1; lptr el,el1; eptr ek,e00,e01,e10,e11;

 if ((!IS_BOUNDING(s,*e,*f))|| IS_INFINITY_FACE(s,*f) || IS_DANGLING_EDGE(s,*e))
  return(FALSE);
 f1=OTHER_FACE(s,*f,*e);
 el=FACE_EDGE(s,*f);
 el1=MARK_POSITION(el);
 EDGE_EDGE(s,*e,&e00,&e01,&e10,&e11);
 do                 /* set face links of edges bounding f */
 {
  ek=(eptr)FIRST(el);
  if (ek!=(*e))
  ek->faces[(*f)==ek->faces[1]]=f1;
  CLOCKWISE(&el);
 } while (!SAME_POSITION(el,el1));
 FREELIST(&el);
          /* set edge links of edges adjacent to e */
 e10->edges[e10->edges[1]==(*e)]=e11;
 e01->edges[e01->edges[1]==(*e)]=e00;
          /* substitute e in edge links of vertices and face */
 if ((*e)->vertices[0]->incidentedge==(*e)) 
  (*e)->vertices[0]->incidentedge=e00;
 if ((*e)->vertices[1]->incidentedge==(*e)) 
  (*e)->vertices[1]->incidentedge=e11;
 if (f1->boundingedge==(*e)) 
  f1->boundingedge=e00;
 f1->rc=0;
 if ((*f)->next!=NULL)              /* discard f and e */
  (*f)->next->prev=(*f)->prev;
 if ((*f)->prev!=NULL)
  (*f)->prev->next=(*f)->next;
 else
  s->flist=(*f)->next;
/* free((char *)(*f)); *f=NULL; */
 free(*f); *f=NULL;
 if ((*e)->next!=NULL) 
  (*e)->next->prev=(*e)->prev;
 if ((*e)->prev!=NULL)
  (*e)->prev->next=(*e)->next;
 else
  s->elist=(*e)->next;
/* free((char *)(*e)); *e=NULL; */
 free(*e); *e=NULL;
 return(TRUE);
}

bool KILL_EDGE_VERTEX(s,e,v)
sptr s; eptr *e; vptr *v;
/* deletes edge e and vertex v from tassellation s */
{
 vptr v1; eptr e00,e01,e10,e11;

 if (!(IS_DANGLING_VERTEX(s,*v) && IS_INCIDENT(s,*e,*v)))
  return(FALSE);
 v1=OTHER_VERTEX(s,*v,*e);
 if (IS_DANGLING_VERTEX(s,v1))                   /* last edge */
 {
  v1->incidentedge=NULL;
  s->infinity->boundingedge=NULL;
 }
 else
 {
  EDGE_EDGE(s,*e,&e00,&e01,&e10,&e11);
  if (*v==(*e)->vertices[0])
  {
   e10->edges[*e==e10->edges[1]]=e11;
   if (v1->incidentedge==(*e))
    v1->incidentedge=e11;
   if (*e==(*e)->faces[0]->boundingedge)
    (*e)->faces[0]->boundingedge=e11;
  }
  else
  {
   e01->edges[*e==e01->edges[1]]=e00;
   if (v1->incidentedge==*e)
    v1->incidentedge=e00;
   if (*e==(*e)->faces[0]->boundingedge)
    (*e)->faces[0]->boundingedge=e00;
  }
 }
 if ((*v)->next!=NULL)              /* discard v and e */
  (*v)->next->prev=(*v)->prev;
 if ((*v)->prev!=NULL) 
  (*v)->prev->next=(*v)->next;
 else
  s->vlist=(*v)->next;
 free(*v); *v=NULL;
 if ((*e)->next!=NULL)
  (*e)->next->prev=(*e)->prev;
 if ((*e)->prev!=NULL)
  (*e)->prev->next=(*e)->next;
 else
  s->elist=(*e)->next;
 free(*e); *e=NULL;
 return(TRUE);
}

bool KILL_VERTEX_INFINITY_FACE(s,v)
/* deletes last vertex v and infinity face from tessellation s */
sptr s; vptr *v;
{
 if (IS_ISOLATED(s,*v))
 {
  free(s->vlist);
  free(s->flist);
  s->vlist=NULL; *v=NULL;
  s->flist=NULL;
  s->infinity=NULL;
  return(TRUE);
 }
 else return(FALSE);
}

bool SWAP_EDGE(s,f,f1,e)
sptr s; fptr *f,*f1; eptr *e;
/* if f and f1 are triangles of s adjacent along e, swap e with the other
   diagonal of the quadrilateral formed by f and f1 */
{
 vptr v1,v2;

 if (IS_A_TRIANGLE(s,f) && IS_A_TRIANGLE(s,f1) &&
     IS_BOUNDING(s,e,f) && IS_BOUNDING(s,e,f1))
 {
  v1=OPPOSITE_VERTEX(s,*f,*e);
  v2=OPPOSITE_VERTEX(s,*f1,*e);
  KILL_EDGE_FACE(s,*e,*f1);
  MAKE_EDGE_FACE(s,f,f1,v1,v2,e);
  return(TRUE);
 }
 else return(FALSE);
}
